package org.ibex.util;
import java.util.*;
import java.io.*;
import java.util.zip.*;
import org.apache.bcel.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.util.*;

        // SecurityManager hacks to avoid java.security?
        // URL and all descendents?  Probably impossible.
        // ObjectInput/ObjectOutput?  Serialization?

        // superprune: URLClassLoader, convert.In/Output other than needed, unneeded locales
        // reflective metadata is killing us...

// Reachability rules:

// - a constructor is reachable iff it is called
// - a static method is reachable iff it is called
// - a nonstatic method is reachable 
// - a static field is reachable iff it is referenced
// - a nonstatic field is reachable iff it is referenced
// - <clinit> is reachable iff any methods, static methods, fields, or constructors are reachable

// - if a method is reachable, all the methods it overrides are reachable

// try eliminating instance methods/fields if no ctor is reachable...
// distinguish between overloaded forms of a given method
// slim down the set of loadAllMethod() classes
// field pruning?

// FIXME: nonstatic method invocation or field access implies that object will be constructed (ie hint)

public class BytecodePruner {

    public static final boolean deleteMethods = false;
    public static SyntheticRepository repo = null;
    public static HashSet dest = new HashSet();
    public static String outdir = ".";

    public void loadAllMethods(String classname) throws Exception {
        visitJavaClass(repo.loadClass(classname));
        Method[] meths = getMethods(repo.loadClass(classname));
        for(int i=0; i<meths.length; i++) visitJavaMethod(repo.loadClass(classname), meths[i]);
    }

    public void loadMethod(String classAndMethodName) throws Exception {
        String classname = classAndMethodName.substring(0, classAndMethodName.lastIndexOf('.'));
        String methodname = classAndMethodName.substring(classAndMethodName.lastIndexOf('.') + 1);
        if (classname.endsWith("." + methodname)) methodname = "<init>";
        visitJavaClass(repo.loadClass(classname));
        Method[] meths = getMethods(repo.loadClass(classname));
        for(int i=0; i<meths.length; i++)
            if (meths[i].getName().equals(methodname))
                visitJavaMethod(repo.loadClass(classname), meths[i]);
    }

    public static void main(String[] s) throws Exception {
        BytecodePruner bcp = new BytecodePruner();
        int start = 1;
        if (s.length >= 3 && s[1].equals("-o")) { outdir = s[2]; start += 2; }
        repo = SyntheticRepository.getInstance(new ClassPath(s[0]));
        for(int i=start; i<s.length; i++)
            if (s[i].endsWith(".class")) bcp.visitJavaClass(repo.loadClass(s[i].substring(0, s[i].length() - 6)));
            else bcp.loadMethod(s[i]);

        bcp.loadMethod("java.lang.Thread.run");                                // we call start(), but the VM calls run()...
        bcp.loadAllMethods("gnu.gcj.runtime.StringBuffer");                    // the compiler emits calls directly to this class
        bcp.loadAllMethods("java.lang.SecurityContext");
        bcp.loadAllMethods("java.lang.ThreadDeath");
        bcp.visitJavaClass(repo.loadClass("java.awt.AWTPermission"));
        bcp.visitJavaClass(repo.loadClass("gnu.classpath.Configuration"));
        bcp.loadAllMethods("java.util.Hashtable$HashIterator");
        bcp.loadMethod("java.util.SimpleTimeZone.useDaylightTime");
        bcp.visitJavaClass(repo.loadClass("gnu.gcj.runtime.FinalizerThread"));
        bcp.visitJavaClass(repo.loadClass("gnu.gcj.runtime.FirstThread"));

        StringTokenizer st = new StringTokenizer(s[0], ":"); 
        while(st.hasMoreTokens()) {
            ZipFile zf = new ZipFile(st.nextToken());
            Enumeration e = zf.entries();
            while(e.hasMoreElements()) {
                String ss = ((ZipEntry)e.nextElement()).getName();
                if (!ss.endsWith(".class")) continue;
                ss = ss.substring(0, ss.length() - 6);
                ss = ss.replace('/', '.');
                dump(repo.loadClass(ss));
            }
        }
    }

    public BytecodePruner() { }

    public static void dump(JavaClass clazz) throws Exception {
        if (clazz.getClassName().startsWith("java.sql.")) return;
        if (!dest.contains(clazz)) return;

        ConstantPoolGen newcpg = new ConstantPoolGen(clazz.getConstantPool());
        ClassGen cg = new ClassGen(clazz);
        cg.setConstantPool(newcpg);
        InstructionFactory factory = new InstructionFactory(cg, newcpg);
        cg.setMajor(46);
        cg.setMinor(0);

        Field[] fields = clazz.getFields();
        int numFields = 0;
        for(int i=0; i<fields.length; i++)
            if (!dest.contains(fields[i]) && false) { 
                System.out.println("  pruning " + clazz.getClassName() + "." + fields[i].getName());
                fields[i] = null; 
            } else numFields++;

        Method[] methods = getMethods(clazz);
        int numMethods = 0;
        boolean good = false;
        for(int i=0; i<methods.length; i++)
            if (clazz.getClassName().startsWith("java.io.FileDescriptor")) {
                good = true;
            } else if (dest.contains(methods[i])) {
                if (!methods[i].getName().equals("<clinit>")) good = true;
            } else {
                if (methods[i].getCode() == null) {
                    System.out.println("  empty codeblock: " + clazz.getClassName() + "." + methods[i].getName());
                } else {
                    System.out.println("  pruning " + clazz.getClassName() + "." + methods[i].getName());
                    if (!deleteMethods) {
                        MethodGen mg = new MethodGen(methods[i], clazz.getClassName(), newcpg);
                        mg.removeExceptions();
                        InstructionList il = new InstructionList();
                        mg.setInstructionList(il);
                        
                        InstructionHandle ih_0 = il.append(factory.createNew("java.lang.UnsatisfiedLinkError"));
                        il.append(InstructionConstants.DUP);
                        il.append(factory.createInvoke("java.lang.UnsatisfiedLinkError",
                                                       "<init>", Type.VOID, Type.NO_ARGS, Constants.INVOKESPECIAL));
                        il.append(InstructionConstants.ATHROW);
                        
                        mg.setMaxStack();
                        mg.setMaxLocals();
                        mg.removeExceptions();
                        mg.removeLocalVariables();
                        mg.removeExceptionHandlers();
                        mg.removeLineNumbers();

                        cg.replaceMethod(methods[i], mg.getMethod());
                        il.dispose();
                    } else {
                        cg.removeMethod(methods[i]);
                    }
                }
            }
        if ((clazz.getClassName().startsWith("gnu.java.locale.LocaleInformation") &&
             !clazz.getClassName().endsWith("LocaleInformation_en") &&
             !clazz.getClassName().endsWith("LocaleInformation") &&
             !clazz.getClassName().endsWith("LocaleInformation_en_US"))
            ||
            ((clazz.getClassName().startsWith("gnu.gcj.convert.Input_") ||
              clazz.getClassName().startsWith("gnu.gcj.convert.Output_")) &&
             !(clazz.getClassName().endsWith("ASCII") ||
               clazz.getClassName().endsWith("UTF8") ||
               clazz.getClassName().endsWith("iconv") ||
               clazz.getClassName().endsWith("8859_1")))
            ||
            (!good &&
             !clazz.isInterface() &&
             !clazz.isAbstract() &&
             !clazz.getClassName().startsWith("java.io.") &&
             !clazz.getClassName().startsWith("java.lang.") &&
             !clazz.getClassName().startsWith("gnu.") &&
             !clazz.getClassName().endsWith("Error") &&
             !clazz.getClassName().endsWith("Exception") &&
             !clazz.getClassName().endsWith("Permission"))) {

            //System.out.println("DROPPING " + clazz.getClassName());
            //return;
        }
        new File(outdir + "/" + new File(clazz.getClassName().replace('.', '/')).getParent()).mkdirs();
        System.out.println("dumping " + clazz.getClassName());
        cg.getJavaClass().dump(outdir + "/" + clazz.getClassName().replace('.', '/') + ".class");
    }

    public JavaClass sig2class(String sig) throws Exception {
        if (sig == null) return null;
        while (sig.length() > 0 && (sig.charAt(0) == 'L' || sig.charAt(0) == '[')) {
            if (sig.charAt(0) == 'L') sig = sig.substring(1, sig.length() - 1);
            else if (sig.charAt(0) == '[') sig = sig.substring(1, sig.length());
        }
        if (sig.length() <= 1) return null;
        if (sig.equals("<null object>")) return null;
        if (sig.startsWith("<return address")) return null;
        return repo.loadClass(sig);
    }
    public void load(String sig) throws Exception {
        if (sig == null) return;
        while (sig.length() > 0 && (sig.charAt(0) == 'L' || sig.charAt(0) == '[')) {
            if (sig.charAt(0) == 'L') sig = sig.substring(1, sig.length() - 1);
            else if (sig.charAt(0) == '[') sig = sig.substring(1, sig.length());
        }
        if (sig.length() <= 1) return;
        if (sig.equals("<null object>")) return;
        if (sig.startsWith("<return address")) return;
        visitJavaClass(repo.loadClass(sig));
    }
    public void load(Type t) throws Exception {
        if (t == null) return;
        //String sig = t.getSignature();
        if (t instanceof ArrayType) load(((ArrayType)t).getElementType());
        if (!(t instanceof ObjectType)) return;
        load(((ObjectType)t).getClassName());
    }

    // hashtable of hashsets
    public static Hashtable subclasses = new Hashtable();

    public String getMethodSignature(Method m, ConstantPoolGen cpg) throws Exception {
        return m.getName() + m.getSignature();
    }

    public String getMethodSignature(InvokeInstruction ii, ConstantPoolGen cpg) throws Exception {
        String sig = "";
        Type[] argtypes = ii.getArgumentTypes(cpg);
        for(int j=0; j<argtypes.length; j++) sig += argtypes[j].getSignature();
        return ii.getMethodName(cpg) + "(" + sig + ")" + ii.getReturnType(cpg).getSignature();
    }

    public static int level = 0;
    public void visitJavaMethod(JavaClass jc, Method method) throws Exception {
        visitJavaClass(jc);
        if (jc.getClassName().equals("java.util.Date") && method.getName().equals("readObject")) return;
        //if (jc.getClassName().equals("java.net.URLClassLoader")) return;
        if (jc.getClassName().indexOf("SharedLib") != -1) return;
        if (jc.getClassName().indexOf("Datagram") != -1) return;
        if (dest.contains(method)) return;
        dest.add(method);
        level += 2;
        for(int i=0; i<level; i++) System.out.print(" ");
        ConstantPool cp = method.getConstantPool();
        ConstantPoolGen cpg = new ConstantPoolGen(cp);
        System.out.println(jc.getClassName() + "." + getMethodSignature(method, cpg));
        markMethodInSubclasses(jc, method, cpg);
        if (method.getCode() == null) { level -= 2; return; }
        byte[] code = method.getCode().getCode();
        InstructionList il = new InstructionList(code);
        Instruction[] instructions = il.getInstructions();
        for(int i=0; i<instructions.length; i++){ 
            Instruction instr = instructions[i];
            if (instr instanceof LoadClass) load(((LoadClass)instr).getLoadClassType(cpg));
            if (instr instanceof CPInstruction) load(((CPInstruction)instr).getType(cpg));
            if (instr instanceof InvokeInstruction) {
                InvokeInstruction ii = (InvokeInstruction)instr;
                String ii_sig = getMethodSignature(ii, cpg);
                JavaClass c = sig2class(ii.getLoadClassType(cpg).getSignature());
                load(ii.getReturnType(cpg));
                load(ii.getType(cpg));
                Method[] meths = getMethods(c);
                boolean good = false;
                for(int i2=0; i2<meths.length; i2++) {
                    if (getMethodSignature(meths[i2], cpg).equals(ii_sig)) {
                        visitJavaMethod(c, meths[i2]);
                        good = true;
                        break;
                    }
                } 
                if (!good) throw new Exception("couldn't find method " + getMethodSignature(ii, cpg) + " in " + c.getClassName());
            }
        }
        level -= 2;
        load(method.getReturnType());
        Type[] argtypes = method.getArgumentTypes();
        for(int i=0; i<argtypes.length; i++) load(argtypes[i]);
        if (method.getExceptionTable() != null) {
            String[] exntypes = method.getExceptionTable().getExceptionNames();
            for(int i=0; i<exntypes.length; i++) load(exntypes[i]);
        }
        if (method.getCode() != null && method.getCode().getExceptionTable() != null) {
            CodeException[] exntypes = method.getCode().getExceptionTable();

            for(int i=0; i<exntypes.length; i++) {
                int index = exntypes[i].getCatchType();
                if (index == 0) continue;
                load(Utility.compactClassName(cp.getConstantString(index, Constants.CONSTANT_Class), false));
            }
        }
    }

    public void visitJavaField(Field field) throws Exception {
        if (dest.contains(field)) return;
        dest.add(field);
        load(field.getType());
    }

    public void visitJavaClass(JavaClass clazz) throws Exception {
        /*
        if (clazz.getClassName().startsWith("java.security.") &&
            !clazz.getClassName().equals("java.security.cert.Certificate")) return;
        */
        if (dest.contains(clazz)) return;
        dest.add(clazz);

        ConstantPoolGen cpg = new ConstantPoolGen(clazz.getConstantPool());
        String name = clazz.getClassName();
        System.out.println(clazz.getClassName() + ".class");

        JavaClass superclass = clazz.getSuperClass();
        for(JavaClass sup = clazz; sup != null; sup = sup.getSuperClass()) {
            if (sup != clazz) {
                if (subclasses.get(sup) == null) subclasses.put(sup, new HashSet());
                ((HashSet)subclasses.get(sup)).add(clazz);
            }
            JavaClass[] interfaces = sup.getAllInterfaces();
            for(int i=0; i<interfaces.length; i++) {
                if (subclasses.get(interfaces[i]) == null) subclasses.put(interfaces[i], new HashSet());
                ((HashSet)subclasses.get(interfaces[i])).add(clazz);
            }
        }

        if (clazz.getClassName().startsWith("org.ibex.")) loadAllMethods(clazz.getClassName());

        for(JavaClass sup = clazz; sup != null; sup = sup.getSuperClass()) {
            if (sup != clazz) {
                visitJavaClass(sup);
                remarkMethods(sup, clazz, cpg);
            }
            JavaClass[] interfaces = sup.getAllInterfaces();
            for(int i=0; i<interfaces.length; i++) {
                visitJavaClass(interfaces[i]);
                remarkMethods(interfaces[i], clazz, cpg);
            }
        }

        Field[] fields = clazz.getFields();
        Method[] methods = getMethods(clazz);
        for(int i=0; i<fields.length; i++) visitJavaField(fields[i]);
        for(int i=0; i<methods.length; i++) {
            if (methods[i].getName().equals("<clinit>")) visitJavaMethod(clazz, methods[i]);

            // only if ctor reachable (?)
            if (methods[i].getName().equals("equals")) visitJavaMethod(clazz, methods[i]);
            if (methods[i].getName().equals("hashCode")) visitJavaMethod(clazz, methods[i]);
            if (methods[i].getName().equals("finalize")) visitJavaMethod(clazz, methods[i]);
            if (methods[i].getName().equals("clone")) visitJavaMethod(clazz, methods[i]);
            if (methods[i].getName().equals("toString")) visitJavaMethod(clazz, methods[i]);
        }
    }

    public void markMethodInSubclasses(JavaClass c, Method m, JavaClass subclass, ConstantPoolGen cpg) throws Exception {
        if (m.isStatic()) return;
        String sig = getMethodSignature(m, cpg);
        Method[] submethods = getMethods(subclass);
        for(int j=0; j<submethods.length; j++)
            //if (getMethodSignature(submethods[j], cpg).equals(sig))
            if (submethods[j].getName().equals(m.getName()))
                visitJavaMethod(subclass, submethods[j]);
    }
    public void markMethodInSubclasses(JavaClass c, Method m, ConstantPoolGen cpg) throws Exception {
        if (m.isStatic()) return;
        HashSet s = (HashSet)subclasses.get(c);
        if (s == null) return;
        Object[] subclasses = s.toArray();
        for(int i=0; i<subclasses.length; i++) {
            JavaClass subclass = (JavaClass)subclasses[i];
            for(int j=0; j<level; j++) System.out.print(" ");
            System.out.println("  [subclass " + subclass.getClassName() + "]");
            if (subclass == c) return;
            markMethodInSubclasses(c, m, subclass, cpg);
        }
    }
        
    public void remarkMethods(JavaClass c, ConstantPoolGen cpg) throws Exception {
        Method[] meths =getMethods(c);
        for(int j=0; j<meths.length; j++) if (dest.contains(meths[j])) markMethodInSubclasses(c, meths[j], cpg);
    }

    public void remarkMethods(JavaClass c, JavaClass target, ConstantPoolGen cpg) throws Exception {
        Method[] meths = getMethods(c);
        for(int j=0; j<meths.length; j++) if (dest.contains(meths[j])) markMethodInSubclasses(c, meths[j], target, cpg);
    }

    public static Hashtable methodsHashtable = new Hashtable();
    public static Method[] getMethods(JavaClass c) {
        Method[] ret = (Method[])methodsHashtable.get(c);
        if (ret == null) methodsHashtable.put(c, ret = c.getMethods());
        return ret;
    }

}
